@@ -86,6 +86,38 @@
 	 buf = <<>>                  :: binary(),
 	 session_id                  :: binary() | undefined}).
 
+-define(opt_map,
+	[{shaper, fun handle_opt/3},
+	 {server_name, fun handle_opt/3},
+	 {turn_ipv4_address, #state.relay_ipv4_ip},
+	 {turn_ipv6_address, #state.relay_ipv6_ip},
+	 {turn_min_port, #state.min_port},
+	 {turn_max_port, #state.max_port},
+	 {turn_max_allocations, #state.max_allocs},
+	 {turn_max_permissions, #state.max_permissions},
+	 {turn_blacklist_clients, #state.blacklist_clients},
+	 {turn_whitelist_clients, #state.whitelist_clients},
+	 {turn_blacklist_peers, #state.blacklist_peers},
+	 {turn_whitelist_peers, #state.whitelist_peers},
+	 {turn_blacklist, #state.blacklist_peers}, % Deprecated.
+	 {turn_whitelist, #state.whitelist_peers}, % Deprecated.
+	 {use_turn, #state.use_turn},
+	 {auth_type, #state.auth},
+	 {auth_realm, #state.realm},
+	 {auth_fun, #state.auth_fun},
+	 {hook_fun, #state.hook_fun},
+	 {inet, none},
+	 {ip, none},
+	 {backlog, none},
+	 {certfile, none},
+	 {dhfile, none},
+	 {ciphers, none},
+	 {protocol_options, none},
+	 {tls, none},
+	 {proxy_protocol, none},
+	 {sock_peer_name, none},
+	 {session_id, none}]).
+
 %%====================================================================
 %% API
 %%====================================================================
@@ -337,7 +369,6 @@ process(State, #stun{class = request,
 	Secret, IsExpired) ->
     Resp = prepare_response(State, Msg),
     AddrPort = State#state.peer,
-    SockMod = State#state.sock_mod,
     case turn_sm:find_allocation(AddrPort) of
 	{ok, Pid} ->
 	    turn:route(Pid, Msg),
@@ -348,6 +379,7 @@ process(State, #stun{class = request,
 	    run_hook(protocol_error, State, R),
 	    send(State, R);
 	_ ->
+	    SockMod = State#state.sock_mod,
 	    Opts = [{sock, State#state.sock},
 		    {sock_mod, SockMod},
 		    {username, Msg#stun.'USERNAME'},
@@ -489,217 +521,55 @@ route_on_turn(State, Msg, Pass) ->
     end.
 
 prepare_state(Opts, Sock, Peer, SockMod) when is_list(Opts) ->
-    ID = case proplists:get_value(session_id, Opts) of
-	     ID0 when is_binary(ID0) ->
-		 ID0; % Stick to listener's session ID.
-	     undefined ->
-		 stun_logger:make_id()
-	 end,
+    ID = get_session_id(Opts),
+    Auth = get_default_auth(Opts),
+    State = #state{session_id = ID,
+		   auth = Auth,
+		   peer = Peer,
+		   sock = Sock,
+		   sock_mod = SockMod},
     stun_logger:set_metadata(stun, SockMod, ID, Peer),
-    case proplists:get_bool(use_turn, Opts) of
-	true ->
-	    lists:foldl(
-	      fun({turn_ip, IP}, State) ->
-		      case prepare_addr(IP) of
-			  {ok, Addr} ->
-			      ?LOG_WARNING("'turn_ip' is deprecated, specify "
-					   "'turn_ipv4_address' and optionally "
-					   "'turn_ipv6_address' instead"),
-			      State#state{relay_ipv4_ip = Addr};
-			  {error, _} ->
-			      ?LOG_ERROR("Wrong 'turn_ip' value: ~p", [IP]),
-			      State
-		      end;
-		 ({turn_ipv4_address, IP}, State) ->
-		      case prepare_addr(IP) of
-			  {ok, Addr} ->
-			      State#state{relay_ipv4_ip = Addr};
-			  {error, _} ->
-			      ?LOG_ERROR("Wrong 'turn_ipv4_address' value: ~p",
-					 [IP]),
-			      State
-		      end;
-		 ({turn_ipv6_address, IP}, State) ->
-		      case prepare_addr(IP) of
-			  {ok, Addr} ->
-			      State#state{relay_ipv6_ip = Addr};
-			  {error, _} ->
-			      ?LOG_ERROR("Wrong 'turn_ipv6_address' value: ~p",
-					 [IP]),
-			      State
-		      end;
-		 ({turn_min_port, Min}, State)
-		    when is_integer(Min), Min > 1024, Min < 65536 ->
-		      State#state{min_port = Min};
-		 ({turn_min_port, Wrong}, State) ->
-		      ?LOG_ERROR("Wrong 'turn_min_port' value: ~p", [Wrong]),
-		      State;
-		 ({turn_max_port, Max}, State)
-		    when is_integer(Max), Max > 1024, Max < 65536 ->
-		      State#state{max_port = Max};
-		 ({turn_max_port, Wrong}, State) ->
-		      ?LOG_ERROR("Wrong 'turn_max_port' value: ~p", [Wrong]),
-		      State;
-		 ({turn_max_allocations, N}, State)
-		    when (is_integer(N) andalso N > 0) orelse is_atom(N) ->
-		      State#state{max_allocs = N};
-		 ({turn_max_allocations, Wrong}, State) ->
-		      ?LOG_ERROR("Wrong 'turn_max_allocations' value: ~p",
-				 [Wrong]),
-		      State;
-		 ({turn_max_permissions, N}, State)
-		    when (is_integer(N) andalso N > 0) orelse is_atom(N) ->
-		      State#state{max_permissions = N};
-		 ({turn_max_permissions, Wrong}, State) ->
-		      ?LOG_ERROR("Wrong 'turn_max_permissions' value: ~p",
-				 [Wrong]),
-		      State;
-		 ({turn_blacklist_clients, B},
-		  #state{blacklist_clients = B0} = State) ->
-		      case lists:all(fun is_valid_subnet/1, B) of
-			  true ->
-			      State#state{blacklist_clients = B0 ++ B};
-			  false ->
-			      ?LOG_ERROR("Wrong 'turn_blacklist_clients' "
-					 "value: ~p", [B]),
-			      State
-		      end;
-		 ({turn_whitelist_clients, W},
-		  #state{whitelist_clients = W0} = State) ->
-		      case lists:all(fun is_valid_subnet/1, W) of
-			  true ->
-			      State#state{whitelist_clients = W0 ++ W};
-			  false ->
-			      ?LOG_ERROR("Wrong 'turn_whitelist_clients' "
-					 "value: ~p", [W]),
-			      State
-		      end;
-		 ({turn_blacklist_peers, B},
-		  #state{blacklist_peers = B0} = State) ->
-		      case lists:all(fun is_valid_subnet/1, B) of
-			  true ->
-			      State#state{blacklist_peers = B0 ++ B};
-			  false ->
-			      ?LOG_ERROR("Wrong 'turn_blacklist_peers' "
-					 "value: ~p", [B]),
-			      State
-		      end;
-		 ({turn_whitelist_peers, W},
-		  #state{whitelist_peers = W0} = State) ->
-		      case lists:all(fun is_valid_subnet/1, W) of
-			  true ->
-			      State#state{whitelist_peers = W0 ++ W};
-			  false ->
-			      ?LOG_ERROR("Wrong 'turn_whitelist_peers' "
-					 "value: ~p", [W]),
-			      State
-		      end;
-		 ({turn_blacklist, B}, #state{blacklist_clients = C,
-					      blacklist_peers = P} = State0) ->
-		      case lists:all(fun is_valid_subnet/1, B) of
-			  true ->
-			      State1 = State0#state{blacklist_clients = C ++ B},
-			      State2 = State1#state{blacklist_peers = P ++ B},
-			      State2;
-			  false ->
-			      ?LOG_ERROR("Wrong 'turn_blacklist' "
-					 "value: ~p", [B]),
-			      State0
-		      end;
-		 ({turn_whitelist, W}, #state{whitelist_clients = C,
-					      whitelist_peers = P} = State0) ->
-		      case lists:all(fun is_valid_subnet/1, W) of
-			  true ->
-			      State1 = State0#state{whitelist_clients = C ++ W},
-			      State2 = State1#state{whitelist_peers = P ++ W},
-			      State2;
-			  false ->
-			      ?LOG_ERROR("Wrong 'turn_whitelist' "
-					 "value: ~p", [W]),
-			      State0
-		      end;
-		 ({shaper, S}, State)
-		    when S == none orelse (is_integer(S) andalso (S > 0)) ->
-		      State#state{shaper = stun_shaper:new(S)};
-		 ({shaper, Wrong}, State) ->
-		      ?LOG_ERROR("Wrong 'shaper' value: ~p", [Wrong]),
-		      State;
-		 ({server_name, none}, State) ->
-		      State#state{server_name = undefined};
-		 ({server_name, S}, State) ->
-		      try
-			  State#state{server_name = iolist_to_binary(S)}
-		      catch _:_ ->
-			      ?LOG_ERROR("Wrong 'server_name' value: ~p", [S]),
-			      State
-		      end;
-		 ({auth_realm, R}, State) ->
-		      try
-			  State#state{realm = iolist_to_binary(R)}
-		      catch _:_ ->
-			      ?LOG_ERROR("Wrong 'auth_realm' value: ~p", [R]),
-			      State
-		      end;
-		 ({auth_fun, F}, State) when is_function(F) ->
-		      State#state{auth_fun = F};
-		 ({auth_fun, Wrong}, State) ->
-		      ?LOG_ERROR("Wrong 'auth_fun' value: ~p", [Wrong]),
-		      State;
-		 ({hook_fun, F}, State) when is_function(F) ->
-		      State#state{hook_fun = F};
-		 ({hook_fun, Wrong}, State) ->
-		      ?LOG_ERROR("Wrong 'hook_fun' value: ~p", [Wrong]),
-		      State;
-		 ({auth_type, anonymous}, State) ->
-		      State#state{auth = anonymous};
-		 ({auth_type, user}, State) ->
-		      State#state{auth = user};
-		 ({auth_type, Wrong}, State) ->
-		      ?LOG_ERROR("Wrong 'auth_type' value: ~p", [Wrong]),
-		      State;
-		 ({use_turn, _}, State) -> State;
-		 (use_turn, State) -> State;
-		 (inet, State) -> State;
-		 ({ip, _}, State) -> State;
-		 ({backlog, _}, State) -> State;
-		 ({certfile, _}, State) -> State;
-		 ({dhfile, _}, State) -> State;
-		 ({ciphers, _}, State) -> State;
-		 ({protocol_options, _}, State) -> State;
-		 ({tls, _}, State) -> State;
-		 (tls, State) -> State;
-		 ({proxy_protocol, _}, State) -> State;
-		 (proxy_protocol, State) -> State;
-		 ({sock_peer_name, _}, State) -> State;
-		 ({session_id, _}, State) -> State;
-		 (Opt, State) ->
-		      ?LOG_ERROR("Ignoring unknown option '~p'", [Opt]),
-		      State
-	      end,
-	      #state{session_id = ID, peer = Peer, sock = Sock,
-		     sock_mod = SockMod, use_turn = true},
-	      Opts);
-	_ ->
-	    #state{session_id = ID, sock = Sock, sock_mod = SockMod,
-		   peer = Peer, hook_fun = proplists:get_value(hook_fun, Opts),
-		   auth = anonymous}
-    end;
+    lists:foldl(
+      fun({Key, Val}, Acc) ->
+	      case proplists:get_value(Key, ?opt_map) of
+		  Pos when is_integer(Pos) ->
+		      setelement(Pos, Acc, Val);
+		  Fun when is_function(Fun) ->
+		      Fun(Key, Val, Acc);
+		  none ->
+		      Acc;
+		  undefined ->
+		      exit({unknown_option, Key})
+	      end
+      end, State, proplists:unfold(Opts));
 prepare_state(State, _Sock, Peer, SockMod) ->
     ID = stun_logger:make_id(),
     stun_logger:set_metadata(stun, SockMod, ID, Peer),
     State#state{session_id = ID, peer = Peer}.
 
-prepare_addr(IPBin) when is_binary(IPBin) ->
-    prepare_addr(binary_to_list(IPBin));
-prepare_addr(IPS) when is_list(IPS) ->
-    inet_parse:address(IPS);
-prepare_addr(T) when is_tuple(T) ->
-    try
-	inet_parse:address(inet_parse:ntoa(T))
-    catch _:_ ->
-	    {error, einval}
+get_default_auth(Opts) ->
+    case proplists:get_bool(use_turn, Opts) of
+	true ->
+	    user;
+	false ->
+	    anonymous
     end.
 
+get_session_id(Opts) ->
+    case proplists:get_value(session_id, Opts) of
+	ID when is_binary(ID) ->
+	    ID; % Stick to listener's session ID.
+	undefined ->
+	    stun_logger:make_id()
+    end.
+
+handle_opt(server_name, none, State) ->
+    State#state{server_name = undefined};
+handle_opt(server_name, Name, State) ->
+    State#state{server_name = Name};
+handle_opt(shaper, Shaper, State) ->
+    State#state{shaper = stun_shaper:new(Shaper)}.
+
 activate_socket(#state{sock = Sock, sock_mod = gen_tcp, shaper = none}) ->
     inet:setopts(Sock, [{active, ?TCP_ACTIVE}]);
 activate_socket(#state{sock = Sock, sock_mod = SockMod, shaper = none}) ->
@@ -780,25 +650,6 @@ unmap_v4_addr({{0, 0, 0, 0, 0, 16#FFFF, D7, D8}, Port}) ->
 unmap_v4_addr(AddrPort) ->
     AddrPort.
 
-is_valid_subnet({{IP1, IP2, IP3, IP4}, Mask}) ->
-    (IP1 >= 0) and (IP1 =< 255) and
-    (IP2 >= 0) and (IP2 =< 255) and
-    (IP3 >= 0) and (IP3 =< 255) and
-    (IP4 >= 0) and (IP4 =< 255) and
-    (Mask >= 0) and (Mask =< 32);
-is_valid_subnet({{IP1, IP2, IP3, IP4, IP5, IP6, IP7, IP8}, Mask}) ->
-    (IP1 >= 0) and (IP1 =< 65535) and
-    (IP2 >= 0) and (IP2 =< 65535) and
-    (IP3 >= 0) and (IP3 =< 65535) and
-    (IP4 >= 0) and (IP4 =< 65535) and
-    (IP5 >= 0) and (IP5 =< 65535) and
-    (IP6 >= 0) and (IP6 =< 65535) and
-    (IP7 >= 0) and (IP7 =< 65535) and
-    (IP8 >= 0) and (IP8 =< 65535) and
-    (Mask >= 0) and (Mask =< 128);
-is_valid_subnet(_) ->
-    false.
-
 get_sockmod(Opts, Sock) ->
     case proplists:get_value(tls, Opts, false) of
 	true ->
